/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
package com.google.code.or;

import com.google.code.or.binlog.BinlogEventV4;
import com.google.code.or.binlog.impl.event.RotateEvent;
import com.google.code.or.io.EOBinlogException;
import com.google.code.or.net.impl.packet.EOFPacket;
import com.google.code.or.net.impl.packet.ResultSetFieldPacket;
import com.google.code.or.net.impl.packet.ResultSetHeaderPacket;
import com.google.code.or.net.impl.packet.ResultSetRowPacket;
import com.google.code.or.net.impl.packet.command.ComQuery;
import com.google.code.or.net.impl.packet.command.ComQuit;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import com.google.code.or.binlog.BinlogEventListener;
import com.google.code.or.binlog.BinlogParser;
import com.google.code.or.binlog.BinlogParserListener;
import com.google.code.or.binlog.impl.ReplicationBasedBinlogParser;
import com.google.code.or.binlog.impl.parser.DeleteRowsEventParser;
import com.google.code.or.binlog.impl.parser.DeleteRowsEventV2Parser;
import com.google.code.or.binlog.impl.parser.FormatDescriptionEventParser;
import com.google.code.or.binlog.impl.parser.IncidentEventParser;
import com.google.code.or.binlog.impl.parser.IntvarEventParser;
import com.google.code.or.binlog.impl.parser.QueryEventParser;
import com.google.code.or.binlog.impl.parser.RandEventParser;
import com.google.code.or.binlog.impl.parser.RotateEventParser;
import com.google.code.or.binlog.impl.parser.StopEventParser;
import com.google.code.or.binlog.impl.parser.TableMapEventParser;
import com.google.code.or.binlog.impl.parser.UpdateRowsEventParser;
import com.google.code.or.binlog.impl.parser.UpdateRowsEventV2Parser;
import com.google.code.or.binlog.impl.parser.UserVarEventParser;
import com.google.code.or.binlog.impl.parser.WriteRowsEventParser;
import com.google.code.or.binlog.impl.parser.WriteRowsEventV2Parser;
import com.google.code.or.binlog.impl.parser.XidEventParser;
import com.google.code.or.common.glossary.column.StringColumn;
import com.google.code.or.io.impl.SocketFactoryImpl;
import com.google.code.or.net.Packet;
import com.google.code.or.net.Transport;
import com.google.code.or.net.TransportException;
import com.google.code.or.net.impl.AuthenticatorImpl;
import com.google.code.or.net.impl.TransportImpl;
import com.google.code.or.net.impl.packet.ErrorPacket;
import com.google.code.or.net.impl.packet.command.ComBinlogDumpPacket;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * @author Jingqi Xu
 *
 * 06.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - fixed a bug related to reset of current position in binlog
 *      - added possibility to set binlog file name and tables for filtering after created parser
 * 16.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - added support of cancel replication via replicator commands interface
 * 19.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - removed handling of exceptions for getBinLogData and showBinaryLogs methods
 *      - added method for setting of waitAvailableTimeout variable
 *      - added updating binlogFilename in parser when handling RotateEvent,
 *        changed mode showing info about RotateEvent from info to debug
 * 24.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - every event now is transferred to outside for handling via callback
 *        (it allows not to create big collections and handle every event at once)
 * 24.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - removed controling of packets count, this is controlled in CDC translator now
 * 17.03.2015 Dmitry Pogorelov DVCORE-2736
 *      - added control of reading packet via waitAvailable method
 */
// EXTENSION
// FIXED BY: Dmitry Pogorelov
// DATE: 2015.01.22
// public class OpenReplicator {
public class OpenReplicator implements OpenReplicatorCommands {
// END EXTENSION
	//
	protected int port = 3306;
	protected String host;
	protected String user;
	protected String password;
	protected int serverId = 6789;
	protected String binlogFileName;
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	protected long binlogPosition;
	// END EXTENSION
	protected String encoding = "utf-8";
	protected int level1BufferSize = 1024 * 1024;
	protected int level2BufferSize = 0;//8 * 1024 * 1024;
	protected int socketReceiveBufferSize = 512 * 1024;
	
	//
	protected Transport transport;
	protected BinlogParser binlogParser;
	protected BinlogEventListener binlogEventListener;
	protected final AtomicBoolean running = new AtomicBoolean(false);
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	protected String checkSum;
	protected List<String> filterTables;

	private static final long START_BINLOG_POSITION = 4;
	private static final Logger LOGGER = LoggerFactory.getLogger(OpenReplicator.class);
	// END EXTENSION

	private Exception parseException;
	private long waitAvailableTimeout = TimeUnit.SECONDS.toMillis(30);
	private Object lock = new Object();


	private final static String SHOW_BINARY_LOGS_COMMAND = "SHOW BINARY LOGS";
	/**
	 * 
	 */
	public boolean isRunning() {
		return this.running.get();
	}
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.21
	public void connect() throws Exception {
		//
		if(this.transport == null) this.transport = getDefaultTransport();
		try {
			this.transport.connect(this.host, this.port, this.waitAvailableTimeout);
		} catch (EOBinlogException e) {
			String errorMessage = "Unexpected error when connecting to MySQL. Try to set a larger waitAvailableTimeout.";
			if (LOGGER.isDebugEnabled())
				LOGGER.debug(errorMessage, e);
			throw new Exception(errorMessage, e);
		}
		//
		setMasterCheckSum();

		//
		initCheckSum();

		//
		if (this.binlogFileName != null && this.binlogFileName.length() > 0) {
			dumpBinlog();
		}
	}

	public void disconnect() {
		//
		this.transport.disconnect();
	}

	private void reconnect() throws Exception {
		disconnect();
		connect();
	}
	// END EXTENSION
	
	public void start() throws Exception {
		//
		if(!this.running.compareAndSet(false, true)) {
			return;
		}
		
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		//if(this.transport == null) this.transport = getDefaultTransport();
		//this.transport.connect(this.host, this.port);

		//
		//dumpBinlog();
		// END EXTENSION
		//
		if(this.binlogParser == null) this.binlogParser = getDefaultBinlogParser();
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		this.binlogParser.setBinlogPosition(this.binlogPosition);
		// END EXTENSION
		this.binlogParser.setWaitAvailableTimeout(this.waitAvailableTimeout);
		this.binlogParser.setBinlogFileName(this.binlogFileName);
		this.binlogParser.setFilterTables(this.filterTables);
		this.binlogParser.setEventListener(this.binlogEventListener);
		this.binlogParser.addParserListener(new BinlogParserListener.Adapter() {
			@Override
			public void onStart(BinlogParser parser) {
				parseException = null;
			}

			@Override
			public void onStop(BinlogParser parser) {
				stopQuietly(0, TimeUnit.MILLISECONDS);
			}
			// EXTENSION
			// FIXED BY: Dmitry Pogorelov
			// DATE: 2015.01.22
			@Override
			public void onException(BinlogParser parser, Exception exception) {
				stopQuietly(0, TimeUnit.MILLISECONDS);
				parseException = exception;
			}
			// END EXTENSION
		});
		this.binlogParser.start();
	}

	public void stop(long timeout, TimeUnit unit) throws Exception {
		//
		if(!this.running.compareAndSet(true, false)) {
			return;
		}
		
		//
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		// this.transport.disconnect();
		// END EXTENSION
		this.binlogParser.stop(timeout, unit);
	}
	
	public void stopQuietly(long timeout, TimeUnit unit) {
		try {
			stop(timeout, unit);
		} catch(Exception e) {
			// NOP
		}
	}
	
	/**
	 * 
	 */
	public int getPort() {
		return port;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public String getHost() {
		return host;
	}

	public void setHost(String host) {
		this.host = host;
	}

	public String getUser() {
		return user;
	}

	public void setUser(String user) {
		this.user = user;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}
	
	public String getEncoding() {
		return encoding;
	}

	public void setEncoding(String encoding) {
		this.encoding = encoding;
	}
	
	public int getServerId() {
		return serverId;
	}

	public void setServerId(int serverId) {
		this.serverId = serverId;
	}
	
	public long getBinlogPosition() {
		return binlogPosition;
	}

	public void setBinlogPosition(long binlogPosition) {
		this.binlogPosition = binlogPosition;
	}
	
	public String getBinlogFileName() {
		return binlogFileName;
	}

	public void setBinlogFileName(String binlogFileName) {
		this.binlogFileName = binlogFileName;
	}
	
	public int getLevel1BufferSize() {
		return level1BufferSize;
	}

	public void setLevel1BufferSize(int level1BufferSize) {
		this.level1BufferSize = level1BufferSize;
	}

	public int getLevel2BufferSize() {
		return level2BufferSize;
	}

	public void setLevel2BufferSize(int level2BufferSize) {
		this.level2BufferSize = level2BufferSize;
	}
	
	public int getSocketReceiveBufferSize() {
		return socketReceiveBufferSize;
	}

	public void setSocketReceiveBufferSize(int socketReceiveBufferSize) {
		this.socketReceiveBufferSize = socketReceiveBufferSize;
	}

	public long getWaitAvailableTimeout() {
		return waitAvailableTimeout;
	}

	public void setWaitAvailableTimeout(long waitAvailableTimeout) {
		this.waitAvailableTimeout = waitAvailableTimeout;
	}

	/**
	 * 
	 */
	public Transport getTransport() {
		return transport;
	}

	public void setTransport(Transport transport) {
		this.transport = transport;
	}
	
	public BinlogParser getBinlogParser() {
		return binlogParser;
	}

	public void setBinlogParser(BinlogParser parser) {
		this.binlogParser = parser;
	}
	
	public BinlogEventListener getBinlogEventListener() {
		return binlogEventListener;
	}

	public void setBinlogEventListener(BinlogEventListener listener) {
		this.binlogEventListener = listener;
	}

	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	public String getCheckSum() {
		return checkSum;
	}

	public void setCheckSum(String checkSum) {
		this.checkSum = checkSum;
	}
	// END EXTENSION

	/**
	 * 
	 */
	protected void dumpBinlog() throws Exception {
		//
		if (getBinlogPosition() <= 0 )
			setBinlogPosition(START_BINLOG_POSITION);

		final ComBinlogDumpPacket command = new ComBinlogDumpPacket();
		command.setBinlogFlag(ComBinlogDumpPacket.BINLOG_DUMP_BLOCK);
		command.setServerId(this.serverId);
		command.setBinlogPosition(this.binlogPosition);
		command.setBinlogFileName(StringColumn.valueOf(this.binlogFileName.getBytes(this.encoding)));
		LOGGER.info("Dumping for binlog file name: " + command.getBinlogFileName());
		this.transport.getOutputStream().writePacket(command);
		this.transport.getOutputStream().flush();
		
		//
		try {
			final Packet packet = this.transport.getInputStream().readPacket();
			if(packet.getPacketBody()[0] == ErrorPacket.PACKET_MARKER) {
				final ErrorPacket error = ErrorPacket.valueOf(packet);
				throw new TransportException(error);
			}
		} catch (EOBinlogException e) {
			String errorMessage = "Unexpected error when dumping of binlog data from MySQL. Try to set a larger waitAvailableTimeout.";
			if (LOGGER.isDebugEnabled())
				LOGGER.debug(errorMessage, e);
			throw new Exception(errorMessage, e);
		}
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		this.transport.getInputStream().setReadLimit(0);
		this.transport.getInputStream().resetPosition();
		// END EXTENSION
	}
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	protected void setMasterCheckSum() throws Exception {
		//
		final ComQuery command = new ComQuery();
		command.setSql(StringColumn.valueOf("SET @master_binlog_checksum = @@global.binlog_checksum".getBytes()));
		transport.getOutputStream().writePacket(command);
		transport.getOutputStream().flush();

		//
		try {
			final Packet packet = this.transport.getInputStream().readPacket();
			if (packet.getPacketBody()[0] == ErrorPacket.PACKET_MARKER) {
				final ErrorPacket error = ErrorPacket.valueOf(packet);
				throw new TransportException("Please check version of MySQL. The binlog_checksum variable was introduced in 5.6.2 version. Error:" + error);
			}
		} catch (EOBinlogException e) {
			String errorMessage = "Unexpected error when setting of master_binlog_checksum value in MySQL. Try to set a larger waitAvailableTimeout.";
			if (LOGGER.isDebugEnabled())
				LOGGER.debug(errorMessage, e);
			throw new Exception(errorMessage, e);
		}
	}
	// EXTENSION
	// ISSUE: DVCORE-2597
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.02.02
	private List<ResultSetRowPacket> sendAndGetResultOf(String sql) throws Exception {
		//
		List<ResultSetRowPacket> result = new ArrayList<ResultSetRowPacket>();

		//
		final ComQuery command = new ComQuery();
		command.setSql(StringColumn.valueOf(sql.getBytes()));
		transport.getOutputStream().writePacket(command);
		transport.getOutputStream().flush();

		//
		Packet packet = this.transport.getInputStream().readPacket();
		if(packet.getPacketBody()[0] == ErrorPacket.PACKET_MARKER) {
			final ErrorPacket error = ErrorPacket.valueOf(packet);
			throw new TransportException(error);
		}

		//
		final ResultSetHeaderPacket header = ResultSetHeaderPacket.valueOf(packet);
		if (LOGGER.isDebugEnabled())
			LOGGER.debug("{}", header);

		//
		while(true) {
			packet = transport.getInputStream().readPacket();
			if(packet.getPacketBody()[0] == EOFPacket.PACKET_MARKER) {
				EOFPacket eof = EOFPacket.valueOf(packet);
				if (LOGGER.isDebugEnabled())
					LOGGER.debug("{}", eof);
				break;
			} else {
				ResultSetFieldPacket field = ResultSetFieldPacket.valueOf(packet);
				if (LOGGER.isDebugEnabled())
					LOGGER.debug("{}", field);
			}
		}

		//
		while(true) {
			packet = transport.getInputStream().readPacket();
			if(packet.getPacketBody()[0] == EOFPacket.PACKET_MARKER) {
				EOFPacket eof = EOFPacket.valueOf(packet);
				if (LOGGER.isDebugEnabled())
					LOGGER.debug("{}", eof);
				break;
			} else {
				ResultSetRowPacket row = ResultSetRowPacket.valueOf(packet);
				if (LOGGER.isDebugEnabled())
					LOGGER.debug("{}", row);
				result.add(row);
			}
		}

		return result;
	}

	private void stopRunningOfCommand() throws Exception {
		//
		final ComQuit command = new ComQuit();
		transport.getOutputStream().writePacket(command);
		transport.getOutputStream().flush();
	}
	// END EXTENSION

	protected void initCheckSum() throws Exception {
		// EXTENSION
		// ISSUE: DVCORE-2597
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.02.02
		try {
			List<ResultSetRowPacket> rows = sendAndGetResultOf("SELECT @@binlog_checksum as checksum");
			if (rows == null || rows.size() == 0) {
				throw new TransportException("Unexpected error while getting of binlog_checksum variable. The binlog_checksum variable was introduced in 5.6.2 version.");
			}
			ResultSetRowPacket result = rows.get(0);
			checkSum = result.getColumns().get(0).toString();
			LOGGER.info("Master binlog is checksummed with : " + checkSum);
		} catch (EOBinlogException e) {
			String errorMessage = "Unexpected error when getting of binlog_checksum value from MySQL. Try to set a larger waitAvailableTimeout.";
			if (LOGGER.isDebugEnabled())
				LOGGER.debug(errorMessage, e);
			throw new Exception(errorMessage, e);
		}
		// END EXTENSION
	}
	// END EXTENSION
	
	protected Transport getDefaultTransport() throws Exception {
		//
		final TransportImpl r = new TransportImpl();
		r.setLevel1BufferSize(this.level1BufferSize);
		r.setLevel2BufferSize(this.level2BufferSize);

		//
		final AuthenticatorImpl authenticator = new AuthenticatorImpl();
		authenticator.setUser(this.user);
		authenticator.setPassword(this.password);
		authenticator.setEncoding(this.encoding);
		r.setAuthenticator(authenticator);
		
		//
		final SocketFactoryImpl socketFactory = new SocketFactoryImpl();
		socketFactory.setKeepAlive(true);
		socketFactory.setTcpNoDelay(false);
		socketFactory.setReceiveBufferSize(this.socketReceiveBufferSize);
		r.setSocketFactory(socketFactory);
		return r;
	}
	
	protected ReplicationBasedBinlogParser getDefaultBinlogParser() throws Exception {
		//
		final ReplicationBasedBinlogParser r = new ReplicationBasedBinlogParser();
		r.registgerEventParser(new StopEventParser());
		r.registgerEventParser(new RotateEventParser());
		r.registgerEventParser(new IntvarEventParser());
		r.registgerEventParser(new XidEventParser());
		r.registgerEventParser(new RandEventParser());
		r.registgerEventParser(new QueryEventParser());
		r.registgerEventParser(new UserVarEventParser());
		//r.registgerEventParser(new IncidentEventParser());
		r.registgerEventParser(new TableMapEventParser());
		r.registgerEventParser(new WriteRowsEventParser());
		r.registgerEventParser(new UpdateRowsEventParser());
		r.registgerEventParser(new DeleteRowsEventParser());
		r.registgerEventParser(new WriteRowsEventV2Parser());
		r.registgerEventParser(new UpdateRowsEventV2Parser());
		r.registgerEventParser(new DeleteRowsEventV2Parser());
		r.registgerEventParser(new FormatDescriptionEventParser());
		
		//
		r.setTransport(this.transport);
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		r.setCheckSum(this.checkSum);
		// END EXTENSION
		return r;
	}

	private void getData(final BinlogEventListener callback, String binlogFileName, long binlogPosition, long waitAvailableTimeout, boolean disconnect) throws Exception {
		if (LOGGER.isDebugEnabled())
			LOGGER.debug("start getBinLogData binlogPosition=" + binlogPosition + " binlogFileName=" + binlogFileName);

		try {
			setWaitAvailableTimeout(waitAvailableTimeout);

			if (getBinlogFileName() == null) {
				setBinlogFileName(binlogFileName);
				setBinlogPosition(binlogPosition);
				try {
					dumpBinlog();
				} catch (SocketTimeoutException e) {
					LOGGER.error("There was an error when dumping. Socket timeout: " + getTransport().getSocketTimeout() + " is expired.", e);
				} catch (Exception e) {
					LOGGER.error("There was an error when dumping", e);
				}
			}

			if (!getBinlogFileName().equalsIgnoreCase(binlogFileName) ||
					(this.binlogParser == null || (this.binlogParser != null && binlogPosition != getBinlogPosition()))
					) {
				if (LOGGER.isDebugEnabled())
					LOGGER.debug("Another binlog file name has been set. Reconnecting to MySQL...");
				setBinlogFileName(binlogFileName);
				setBinlogPosition(binlogPosition);
				reconnect();
			}

			if (!this.transport.checkSocket()) {
				LOGGER.info("There was a problem with socket. Probably the socket was closed. Trying to reconnect to MySQL...");
				reconnect();
			}

			setBinlogEventListener(new BinlogEventListener() {
				public void onEvents(BinlogEventV4 event) {
					Transport transport = getTransport();
					if (event instanceof RotateEvent) {
						if (LOGGER.isDebugEnabled())
							LOGGER.debug("There was Rotate Event. Handling...");
						RotateEvent rotateEvent = (RotateEvent) event;
						String newBinlogFileName = rotateEvent.getBinlogFileName().toString();

						transport.getInputStream().resetPosition();
						setBinlogFileName(newBinlogFileName);
						binlogParser.setBinlogFileName(newBinlogFileName);
					}
					setBinlogPosition(event.getHeader().getNextBinlogPosition());
					if (callback != null) callback.onEvents(event);
				}
			});

			start();

			if (parseException != null) throw new Exception(parseException);
		} catch (Exception e) {
			disconnect();
			throw e;
		} finally {
			if (disconnect) disconnect();
		}
	}

	@Override
	// EXTENSION
	// ISSUE: DVCORE-2597
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.02.04
	//public List<BinlogEventV4> getBinLogData(String binlog, long binlogPosition, String tables) {
	public void getBinLogData(BinlogEventListener callback,
	                          String binlog,
	                          long binlogPosition,
	                          String tables,
	                          long waitAvailableTimeout,
	                          boolean disconnect) throws Exception {
	// END EXTENSION
		synchronized (lock) {
			setTableFilter(tables);
			// EXTENSION
			// ISSUE: DVCORE-2597
			// FIXED BY: Dmitry Pogorelov
			// DATE: 2015.02.04
			//return getData(binlog, binlogPosition, 0);
			getData(callback, binlog, binlogPosition, waitAvailableTimeout, disconnect);
			// END EXTENSION
		}
	}

	// EXTENSION
	// ISSUE: DVCORE-2597
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.02.02
	@Override
	public List<List<?>> showBinaryLogs() throws Exception {
		synchronized (lock) {
			List<List<?>> result = new ArrayList<List<?>>();
			if (getBinlogFileName() != null) {
				setBinlogFileName(null);
				reconnect();
			}

			if (!this.transport.checkSocket()) {
				LOGGER.info("There was a problem with socket. Probably the socket was closed. Trying to reconnect to MySQL...");
				reconnect();
			}

			try {
				List<ResultSetRowPacket> binaryRows;
				try {
					binaryRows = sendAndGetResultOf(SHOW_BINARY_LOGS_COMMAND);
				} catch (SocketException e) {
					reconnect();
					binaryRows = sendAndGetResultOf(SHOW_BINARY_LOGS_COMMAND);
				}
				for (ResultSetRowPacket binaryRow : binaryRows) {
					List<StringColumn> cols = binaryRow.getColumns();
					List<Object> row = new ArrayList<Object>();
					row.add(cols.get(0).toString());
					row.add(cols.get(1).toString());
					result.add(row);
				}
			} catch (EOBinlogException e) {
				String errorMessage = "Unexpected error when getting of binary logs list from MySQL. Try to set a larger waitAvailableTimeout.";
				if (LOGGER.isDebugEnabled())
					LOGGER.debug(errorMessage, e);
				throw new Exception(errorMessage, e);
			} catch (Exception e) {
				throw new Exception("Unexpected error when getting of binary logs list", e);
			}

			return result;
		}
	}
	// END EXTENSION

	private void setTableFilter(String tables) {
		if (tables != null && tables.length() > 0) {
			String[] parsedTables = tables.split(",");
			this.filterTables = new ArrayList<String>();
			for (String table : parsedTables) {
				this.filterTables.add(table.toLowerCase().trim());
			}
		} else {
			this.filterTables = null;
		}
	}
	// END EXTENSION


	@Override
	public void stopReplication() {
		stopQuietly(0, TimeUnit.MILLISECONDS);
	}
}
