/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
package com.google.code.or.binlog.impl;

import com.google.code.or.common.util.MySQLConstants;
import com.google.code.or.common.util.MySQLUtils;
import com.google.code.or.io.EOBinlogException;
import java.io.IOException;
import java.util.concurrent.TimeUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import com.google.code.or.binlog.BinlogEventParser;
import com.google.code.or.binlog.impl.event.BinlogEventV4HeaderImpl;
import com.google.code.or.io.XInputStream;
import com.google.code.or.net.Transport;
import com.google.code.or.net.impl.packet.EOFPacket;
import com.google.code.or.net.impl.packet.ErrorPacket;
import com.google.code.or.net.impl.packet.OKPacket;

/**
 * 
 * @author Jingqi Xu
 *
 * changes:
 * 04.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - TABLE_MAP_EVENT with row event go as single packet
 * 19.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - moved logging of current position from XInputStream for adding binlogFilename
 * 24.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - removed controling of packets count, this is controlled in CDC translator now
 */
public class ReplicationBasedBinlogParser extends AbstractBinlogParser {
	//
	private static final Logger LOGGER = LoggerFactory.getLogger(ReplicationBasedBinlogParser.class);
	
	//
	protected Transport transport;
	protected String binlogFileName;

	/**
	 * 
	 */
	public ReplicationBasedBinlogParser() {
	}
	
	@Override
	protected void doStart() throws Exception {
		// NOP
	}

	@Override
	protected void doStop(long timeout, TimeUnit unit) throws Exception {
		// NOP
	}
	
	/**
	 * 
	 */
	public Transport getTransport() {
		return transport;
	}

	public void setTransport(Transport transport) {
		this.transport = transport;
	}
	
	public String getBinlogFileName() {
		return binlogFileName;
	}

	public void setBinlogFileName(String binlogFileName) {
		this.binlogFileName = binlogFileName;
	}

	/**
	 * 
	 */
	@Override
	protected void doParse() throws Exception {
		//
		final XInputStream is = this.transport.getInputStream();
		is.setTimeoutMillis(this.waitAvailableTimeout);

		final Context context = new Context(this.binlogFileName, this.checkSum);
		context.setFilterTables(getFilterTables());
		long currentPosition = this.binlogPosition;
		//
		while(isRunning() && this.transport.checkSocket() && is.waitAvailable() > 0) {
			long position = is.getCurrentPosition();
			try {
				// Parse packet
				final int packetLength = is.readInt(3);
				final int packetSequence = is.readInt(1);
				setReadLimit(is, packetLength, packetSequence); // Ensure the packet boundary

				//
				final int packetMarker = is.readInt(1);
				if (packetMarker != OKPacket.PACKET_MARKER) { // 0x00
					if ((byte) packetMarker == ErrorPacket.PACKET_MARKER) { // 0xFF
						final ErrorPacket packet = ErrorPacket.valueOf(packetLength, packetSequence, packetMarker, is);
						throw new RuntimeException(packet.toString() + " BinlogFilename: " +this.binlogFileName + ", position: " + position + ".");
					} else if ((byte) packetMarker == EOFPacket.PACKET_MARKER) { // 0xFE
						throw new EOBinlogException(EOBinlogException.EOF_MARKER);
					} else {
						throw new RuntimeException("Assertion failed, invalid packet marker: " + packetMarker + ". Maybe position is incorrect." +
								" BinlogFilename: " +this.binlogFileName + ", position: " + position + ".");
					}
				}

				// Parse the event header
				final BinlogEventV4HeaderImpl header = new BinlogEventV4HeaderImpl();
				header.setCurrentBinlogPosition(currentPosition);
				header.setTimestamp(is.readLong(4) * 1000L);
				header.setEventType(is.readInt(1));
				header.setServerId(is.readLong(4));
				header.setEventLength(is.readLong(4));
				header.setNextPosition(is.readLong(4));
				header.setFlags(is.readInt(2));
				header.setTimestampOfReceipt(System.currentTimeMillis());
				int eventType = header.getEventType();

				if (header.getNextPosition() == 0 && eventType != MySQLConstants.ROTATE_EVENT)
					// for example FormatDescriptionEvent has nextPosition=0 meaning that this event is a common part of the next event
					header.setNextBinlogPosition(currentPosition);
				else
					header.setNextBinlogPosition(header.getNextPosition());

				if (isVerbose() && LOGGER.isDebugEnabled()) {
					LOGGER.debug("Received an event, sequence: {}, header: {}", packetSequence, header);
				}

				// Parse the event body
				if (this.eventFilter != null && !this.eventFilter.accepts(header, context)) {
					this.defaultParser.parse(is, header, context);
				} else {
					BinlogEventParser parser = getEventParser(eventType);
					if (parser == null) parser = this.defaultParser;
					parser.parse(is, header, context);
				}

				// Ensure the packet boundary
				if (is.available() != 0) {
					if (MySQLUtils.isChecksumUsed(checkSum)) {
						is.skip(4);
						//LOGGER.info("4 bytes were skipped, it seemed to be CRC32 checksum");
						if (is.available() != 0)
							throw new RuntimeException("Assertion failed, available: " + is.available() + ", event type: " + eventType +
									", binlogFilename: " +this.binlogFileName + ", position: " + position + ".");
					} else
						throw new RuntimeException("Assertion failed, available: " + is.available() + ", event type: " + eventType +
								", binlogFilename: " +this.binlogFileName + ", position: " + position + ".");
				}
				currentPosition = header.getNextBinlogPosition();
			} catch (EOBinlogException e) {
				// if the end has been reached
				String errorMessage = "The end of binlog has been reached when parsing packet. There are no available bytes." +
						" Reading of the binlogFilename: " + this.binlogFileName + " position: " + position + " binlog will be continued later." +
						" Maybe value of waitAvailableTime is not enough.";
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug(errorMessage, e);
				}
				stop(0, TimeUnit.MILLISECONDS);
			} finally {
				setReadLimit(is, 0, -1);
				// just for info
				try {
					if (is.waitAvailable() <= 0) {
						if (LOGGER.isDebugEnabled()) {
							LOGGER.debug("The end of binlog has been reached. There are no available bytes." +
									" Reading of the binlogFilename: " + this.binlogFileName + " position: " + is.getCurrentPosition() + " is stopped.");
						}
					}
				} catch (IOException e) {
					if (LOGGER.isDebugEnabled()) {
						LOGGER.debug("Unexpected error appeared when waiting for available data in socket, the socket could be already closed", e);
					}
				}
			}
		}
	}

	private void setReadLimit(XInputStream is, int limit, int packetNumber) throws IOException {
		is.setReadLimit(limit, packetNumber);
		if (LOGGER.isTraceEnabled())
			LOGGER.trace("BinlogFilename: " + binlogFileName + " position: " + is.getCurrentPosition() + ".");
	}
}
