/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
package com.google.code.or.binlog.impl.parser;

import com.google.code.or.common.glossary.OptionalMetadata;
import com.google.code.or.common.util.MySQLUtils;
import java.io.IOException;

import com.google.code.or.binlog.BinlogEventV4Header;
import com.google.code.or.binlog.BinlogParserContext;
import com.google.code.or.binlog.impl.event.TableMapEvent;
import com.google.code.or.common.glossary.Metadata;
import com.google.code.or.io.XInputStream;

/**
 * 
 * @author Jingqi Xu
 */
public class TableMapEventParser extends AbstractBinlogEventParser {
	// 
	private boolean reusePreviousEvent = true;

	/**
	 * 
	 */
	public TableMapEventParser() {
		super(TableMapEvent.EVENT_TYPE);
	}
	
	/**
	 * 
	 */
	public boolean isReusePreviousEvent() {
		return reusePreviousEvent;
	}

	public void setReusePreviousEvent(boolean reusePreviousEvent) {
		this.reusePreviousEvent = reusePreviousEvent;
	}
	
	/**
	 * 
	 */
	public void parse(XInputStream is, BinlogEventV4Header header, BinlogParserContext context)
	throws IOException {
		//
		final long tableId = is.readLong(6); // table Id
		if(this.reusePreviousEvent && context.getTableMapEvent(tableId) != null) {
			is.skip(is.available());
			final TableMapEvent event = context.getTableMapEvent(tableId).copy();
			event.setHeader(header);
			context.getEventListener().onEvents(event);
			return;
		}
		
		//
		final TableMapEvent event = new TableMapEvent(header);
		event.setTableId(tableId);
		event.setReserved(is.readInt(2)); // flags
		event.setDatabaseNameLength(is.readInt(1)); // schema name length
		event.setDatabaseName(is.readNullTerminatedString()); // schema name
		event.setTableNameLength(is.readInt(1)); // table name length
		event.setTableName(is.readNullTerminatedString()); // table name
		event.setColumnCount(is.readUnsignedLong());  // column-count
		event.setColumnTypes(is.readBytes(event.getColumnCount().intValue())); // [length=$column-count] column-def
		event.setColumnMetadataCount(is.readUnsignedLong());  // column-meta-def-count
		event.setColumnMetadata(Metadata.valueOf(event.getColumnTypes(), is.readBytes(event.getColumnMetadataCount().intValue()))); // column-meta-def
		event.setColumnNullabilities(is.readBit(event.getColumnCount().intValue()));

		// some additional metadata info was introduced in MySQL 8, https://mysqlhighavailability.com/more-metadata-is-written-into-binary-log/
		if (is.available() > 0) {
			event.setOptionalMetadata(OptionalMetadata.valueOf(event.getColumnTypes(), is, MySQLUtils.isChecksumUsed(context.getCheckSum()) ? 4 : 0));
		}
		context.getEventListener().onEvents(event);
	}
}
