/**
 * Copyright © 2012-2020 by Data Virtuality GmbH
 */
package com.google.code.or.common.glossary;

import com.google.code.or.common.glossary.column.BitColumn;
import com.google.code.or.common.util.CodecUtils;
import com.google.code.or.common.util.MySQLConstants;
import com.google.code.or.io.XInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * @author Dmitry Pogorelov
 * @date 2020.05.01
 */
public class OptionalMetadata implements Serializable {

	private static final long serialVersionUID = -1427018358281265959L;

	private BitColumn signedness;
	private DefaultCharset defaultCharset;
	private List<Integer> columnCharsets;
	private List<String> columnNames;
	private List<List<String>> setStrValues;
	private List<List<String>> enumStrValues;
	private List<Integer> geometryTypes;
	private List<Integer> simplePrimaryKeys;
	private Map<Integer, Integer> primaryKeysWithPrefix;
	private DefaultCharset enumAndSetDefaultCharset;
	private List<Integer> enumAndSetColumnCharsets;

	public static final OptionalMetadata valueOf(byte[] types, XInputStream is, int skipBytes)
			throws IOException {
		OptionalMetadata result = new OptionalMetadata();
		// one cycle reads one TLV
		while ( is.available() - skipBytes > 0) {
			OptionalFieldType optionalFieldType = OptionalFieldType.byCode(is.readPackedInteger());
			int length = is.readPackedInteger();

			switch (optionalFieldType) {
				case SIGNEDNESS:
					int numericColumnCount = numericColumnCount(types);
					result.setSignedness(readBit(is, numericColumnCount));
					break;
				case DEFAULT_CHARSET:
					result.setDefaultCharset(readDefaultCharset(is, length));
					break;
				case COLUMN_CHARSET:
					result.setColumnCharsets(readIntegers(is, length));
					break;
				case COLUMN_NAME:
					result.setColumnNames(readStrings(is, length));
					break;
				case SET_STR_VALUE:
					result.setSetStrValues(readSetStrings(is, length));
					break;
				case ENUM_STR_VALUE:
					result.setEnumStrValues(readSetStrings(is, length));
					break;
				case GEOMETRY_TYPE:
					result.setGeometryTypes(readIntegers(is, length));
					break;
				case SIMPLE_PRIMARY_KEY:
					result.setSimplePrimaryKeys(readIntegers(is, length));
					break;
				case PRIMARY_KEY_WITH_PREFIX:
					result.setPrimaryKeysWithPrefix(readIntegerPairs(is, length));
					break;
				case ENUM_AND_SET_DEFAULT_CHARSET:
					result.setEnumAndSetDefaultCharset(readDefaultCharset(is, length));
					break;
				case ENUM_AND_SET_COLUMN_CHARSET:
					result.setEnumAndSetColumnCharsets(readIntegers(is, length));
					break;
				default:
					throw new IOException("Unsupported table metadata field type " + optionalFieldType);
			}
		}
		return null;
	}

	private static BitColumn readBit(XInputStream is, int length) throws IOException {
		return is.readBit(length);
	}

	private static List<String> readStrings(XInputStream is, int length) throws IOException {
		List<String> result = new ArrayList<>();
		int i = length;
		while (i > 0) {
			int before = is.available();
			result.add(is.readLengthCodedString().toString());
			i = i - (before - is.available());
		}
		return result;
	}

	private static List<Integer> readIntegers(XInputStream is, int length) throws IOException {
		List<Integer> result = new ArrayList<>();
		int i = length;
		while (i > 0) {
			int before = is.available();
			result.add(is.readPackedInteger());
			i = i - (before - is.available());
		}
		return result;
	}

	private static Map<Integer, Integer> readIntegerPairs(XInputStream is, int length) throws IOException {
		Map<Integer, Integer> result = new HashMap<>();
		int i = length;
		while (i > 0) {
			int before = is.available();
			int columnIndex = is.readPackedInteger();
			int columnCharset = is.readPackedInteger();
			result.put(columnIndex, columnCharset);
			i = i - (before - is.available());
		}
		return result;
	}

	private static List<List<String>> readSetStrings(XInputStream is, int length) throws IOException {
		List<List<String>> result = new ArrayList<>();
		int i = length;
		while (i > 0) {
			int before = is.available();
			int valuesCount = is.readPackedInteger();
			i = i - (before - is.available());
			List<String> typeValues = new ArrayList<>();
			for (int c = 0; c < valuesCount; ++c) {
				before = is.available();
				typeValues.add(is.readLengthCodedString().toString());
				i = i - (before - is.available());
			}
			result.add(typeValues);
		}
		return result;
	}

	private static DefaultCharset readDefaultCharset(XInputStream is, int length) throws IOException {
		DefaultCharset defaultCharset = new DefaultCharset();
		int before = is.available();
		defaultCharset.setDefaultCharsetCollation(is.readPackedInteger());
		Map<Integer, Integer> charsetCollations = readIntegerPairs(is, length - (before - is.available()));
		if (!charsetCollations.isEmpty()) {
			defaultCharset.setCharsetCollations(charsetCollations);
		}

		return defaultCharset;
	}

	private static int numericColumnCount(byte[] types) {
		int count = 0;
		for ( int i = 0 ; i < types.length; i++) {
			final int t = CodecUtils.toUnsigned(types[i]);
			switch (t) {
				case MySQLConstants.TYPE_TINY:
				case MySQLConstants.TYPE_SHORT:
				case MySQLConstants.TYPE_INT24:
				case MySQLConstants.TYPE_LONG:
				case MySQLConstants.TYPE_LONGLONG:
				case MySQLConstants.TYPE_NEWDECIMAL:
				case MySQLConstants.TYPE_FLOAT:
				case MySQLConstants.TYPE_DOUBLE:
					count++;
			}
		}
		return count;
	}

	public BitColumn getSignedness() {
		return signedness;
	}

	public void setSignedness(BitColumn signedness) {
		this.signedness = signedness;
	}

	public DefaultCharset getDefaultCharset() {
		return defaultCharset;
	}

	public void setDefaultCharset(DefaultCharset defaultCharset) {
		this.defaultCharset = defaultCharset;
	}

	public List<Integer> getColumnCharsets() {
		return columnCharsets;
	}

	public void setColumnCharsets(List<Integer> columnCharsets) {
		this.columnCharsets = columnCharsets;
	}

	public List<String> getColumnNames() {
		return columnNames;
	}

	public void setColumnNames(List<String> columnNames) {
		this.columnNames = columnNames;
	}

	public List<List<String>> getSetStrValues() {
		return setStrValues;
	}

	public void setSetStrValues(List<List<String>> setStrValues) {
		this.setStrValues = setStrValues;
	}

	public List<List<String>> getEnumStrValues() {
		return enumStrValues;
	}

	public void setEnumStrValues(List<List<String>> enumStrValues) {
		this.enumStrValues = enumStrValues;
	}

	public List<Integer> getGeometryTypes() {
		return geometryTypes;
	}

	public void setGeometryTypes(List<Integer> geometryTypes) {
		this.geometryTypes = geometryTypes;
	}

	public List<Integer> getSimplePrimaryKeys() {
		return simplePrimaryKeys;
	}

	public void setSimplePrimaryKeys(List<Integer> simplePrimaryKeys) {
		this.simplePrimaryKeys = simplePrimaryKeys;
	}

	public Map<Integer, Integer> getPrimaryKeysWithPrefix() {
		return primaryKeysWithPrefix;
	}

	public void setPrimaryKeysWithPrefix(Map<Integer, Integer> primaryKeysWithPrefix) {
		this.primaryKeysWithPrefix = primaryKeysWithPrefix;
	}

	public DefaultCharset getEnumAndSetDefaultCharset() {
		return enumAndSetDefaultCharset;
	}

	public void setEnumAndSetDefaultCharset(DefaultCharset enumAndSetDefaultCharset) {
		this.enumAndSetDefaultCharset = enumAndSetDefaultCharset;
	}

	public List<Integer> getEnumAndSetColumnCharsets() {
		return enumAndSetColumnCharsets;
	}

	public void setEnumAndSetColumnCharsets(List<Integer> enumAndSetColumnCharsets) {
		this.enumAndSetColumnCharsets = enumAndSetColumnCharsets;
	}

	public static class DefaultCharset implements Serializable {
		private long defaultCharsetCollation;
		private Map<Integer, Integer> charsetCollations;

		public void setDefaultCharsetCollation(long defaultCharsetCollation) {
			this.defaultCharsetCollation = defaultCharsetCollation;
		}

		public long getDefaultCharsetCollation() {
			return defaultCharsetCollation;
		}

		public void setCharsetCollations(Map<Integer, Integer> charsetCollations) {
			this.charsetCollations = charsetCollations;
		}

		public Map<Integer, Integer> getCharsetCollations() {
			return charsetCollations;
		}

		@Override
		public String toString() {
			return "DefaultCharset{" +
					"defaultCharsetCollation=" + defaultCharsetCollation +
					", charsetCollations=" + charsetCollations +
					'}';
		}
	}

	@Override
	public String toString() {
		return "OptionalMetadata{" +
				"signedness=" + signedness +
				", defaultCharset=" + defaultCharset +
				", columnCharsets=" + columnCharsets +
				", columnNames=" + columnNames +
				", setStrValues=" + setStrValues +
				", enumStrValues=" + enumStrValues +
				", geometryTypes=" + geometryTypes +
				", simplePrimaryKeys=" + simplePrimaryKeys +
				", primaryKeysWithPrefix=" + primaryKeysWithPrefix +
				", enumAndSetDefaultCharset=" + enumAndSetDefaultCharset +
				", enumAndSetColumnCharsets=" + enumAndSetColumnCharsets +
				'}';
	}
}
