/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * Modified by
 * @author  Data Virtuality GmbH
 * @date    $Date$
 *  
 */
package com.google.code.or.io.impl;

import com.google.code.or.io.EOBinlogException;
import com.google.code.or.net.impl.TransportImpl;
import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

import com.google.code.or.common.glossary.UnsignedLong;
import com.google.code.or.common.glossary.column.BitColumn;
import com.google.code.or.common.glossary.column.StringColumn;
import com.google.code.or.common.util.CodecUtils;
import com.google.code.or.io.ExceedLimitException;
import com.google.code.or.io.XInputStream;
import com.google.code.or.io.util.XSerializer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * 
 * @author Jingqi Xu
 *
 * 06.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - fixed a bug related to wrong result of available method (because of buffer)
 * 19.02.2015 Dmitry Pogorelov DVCORE-2597
 *      - added support of waiting for available bytes from stream by timeout to avoid blocked reading
 *      - moved logging of current position to parser for adding binlogFilename
 *      - added boolean waitAvailable method for checking whether the end of binlog is reached
 */
public class XInputStreamImpl extends InputStream implements XInputStream {
	//
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	private static final Logger LOGGER = LoggerFactory.getLogger(XInputStreamImpl.class);
	// END EXTENSION
	//
	private int head = 0;
	private int tail = 0;
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	private long currentPosition = 0;
	// END EXTENSION
	private int readCount = 0;
	private int readLimit = 0;
	private int packetNumber = -1;
	private long timeoutMillis = 0;
	private final byte[] buffer;
	private final InputStream is;
	private TransportImpl.TransportCallback transportCallback;

	/**
	 * 
	 */
	public XInputStreamImpl(InputStream is) {
		this(is, 512 *  1024, null);
	}
	
	public XInputStreamImpl(InputStream is, int size, TransportImpl.TransportCallback transportCallback) {
		this.is = is;
		this.buffer = new byte[size];
		this.transportCallback = transportCallback;
	}

	/**
	 * 
	 */
	public int readInt(int length) throws IOException {
		return readInt(length, true);
	}
	
	public long readLong(int length) throws IOException {
		return readLong(length, true);
	}
	
	public byte[] readBytes(int length) throws IOException {
		final byte[] r = new byte[length];
		this.read(r, 0, length);
		return r;
	}
	
	public BitColumn readBit(int length) throws IOException {
		return readBit(length, true);
	}
	
	public UnsignedLong readUnsignedLong() throws IOException {
		final int v = this.read();
		if(v < 251) return UnsignedLong.valueOf(v);
		else if(v == 251) return null;
		else if(v == 252) return UnsignedLong.valueOf(readInt(2));
		else if(v == 253) return UnsignedLong.valueOf(readInt(3));
		else if(v == 254) return UnsignedLong.valueOf(readLong(8));
		else throw new RuntimeException("assertion failed, should NOT reach here");
	}

	public int readPackedInteger() throws IOException {
		Number number = readUnsignedLong();
		if (number == null) {
			throw new IOException("Unexpected NULL where int should have been");
		}
		if (number.longValue() > Integer.MAX_VALUE) {
			throw new IOException("Stumbled upon long even though int expected");
		}
		return number.intValue();
	}
	
	public StringColumn readLengthCodedString() throws IOException {
		final UnsignedLong length = readUnsignedLong();
		return length == null ? null : readFixedLengthString(length.intValue()); 
	}
	
	public StringColumn readNullTerminatedString() throws IOException {
		final XSerializer s = new XSerializer(128); // 128 should be OK for most schema names
		while(true) {
			final int v = this.read();
			if(v == 0) break;
			s.writeInt(v, 1);
		}
		return StringColumn.valueOf(s.toByteArray());
	}
	
	public StringColumn readFixedLengthString(final int length) throws IOException {
		return StringColumn.valueOf(readBytes(length));
	}

	/**
	 * 
	 */
	public int readInt(int length, boolean littleEndian) throws IOException {
		int r = 0;
		for(int i = 0; i < length; ++i) {
			final int v = this.read();
			if(littleEndian) {
				r |= (v << (i << 3));
			} else {
				r = (r << 8) | v;
			}
		}
		return r;
	}

	public long readLong(int length, boolean littleEndian) throws IOException {
		long r = 0;
		for(int i = 0; i < length; ++i) {
			final long v = (long)this.read();
			if(littleEndian) {
				r |= (v << (i << 3));
			} else {
				r = (r << 8) | v;
			}
		}
		return r;
	}
	
	public BitColumn readBit(int length, boolean littleEndian) throws IOException {
		byte[] bytes = readBytes((int)((length + 7) >> 3));
		if(!littleEndian) bytes = CodecUtils.toBigEndian(bytes);
		return BitColumn.valueOf(length, bytes);
	}
	
	/**
	 * 
	 */
	@Override
	public void close() throws IOException {
		this.is.close();
	}

	public void setReadLimit(int limit) {
		setReadLimit(limit, -1);
	}

	public void setReadLimit(int limit, int packetNumber) {
		// EXTENSION
		// FIXED BY: Dmitry Pogorelov
		// DATE: 2015.01.22
		this.currentPosition += this.readCount;
		// END EXTENSION
		this.readCount = 0;
		this.readLimit = limit;
		this.packetNumber = packetNumber;
	}
	
	@Override
	public int available() throws IOException {
		if(this.readLimit > 0) {
			return this.readLimit - this.readCount;
		} else {
			return this.tail - this.head + this.is.available();
		}
	}
	
	public boolean hasMore() throws IOException {
		if(this.head < this.tail) return true;
		return this.available() > 0;
	}
	
	@Override
	public long skip(final long n) throws IOException {
		if(this.readLimit > 0 && (this.readCount + n) > this.readLimit) {
			this.readCount += doSkip(this.readLimit - this.readCount, this.readLimit <= 0);
			throw new ExceedLimitException();
		} else {
			this.readCount += doSkip(n, this.readLimit <= 0);
			return n; // always skip the number of bytes specified by parameter "n"
		}
	}
	
	@Override
	public int read() throws IOException {
		if(this.readLimit > 0 && (this.readCount + 1) > this.readLimit) {
			throw new ExceedLimitException();
		} else {
			if(this.head >= this.tail) doFill();
			final int r = this.buffer[this.head++] & 0xff;
			++this.readCount;
			return r;
		}
	}
	
	@Override
	public int read(final byte b[], final int off, final int len) throws IOException {
		if(this.readLimit > 0 && (this.readCount + len) > this.readLimit) {
			this.readCount += doRead(b, off, this.readLimit - this.readCount);
			throw new ExceedLimitException();
		} else {
			this.readCount += doRead(b, off, len);
			return len; // always read the number of bytes specified by parameter "len"
		}
	}
	// EXTENSION
	// FIXED BY: Dmitry Pogorelov
	// DATE: 2015.01.22
	@Override
	public void setCurrentPosition(long position) throws IOException {
		if (position <= 0) return;
		if (position > this.currentPosition) {
			try {
				this.currentPosition += doSkip(position - this.currentPosition, true);
			} catch (EOBinlogException e) {
				String message = "End of stream has been reached when positioning. Either binlog position or binlog filename is incorrect";
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug(message, e);
				}
				throw new IOException(message, e);
			}
		} else if (position < this.currentPosition) {
			if (is.markSupported()) {
				this.currentPosition = 0;
				this.is.reset();
				doFill();
				doSkip(position, true);
			} else {
				String message = "New position: " + position + " is less than current position: " + this.currentPosition;
				if (LOGGER.isDebugEnabled()) {
					LOGGER.debug(message);
				}
				throw new IOException(message);
			}
		}
	}

	@Override
	public long getCurrentPosition() {
		return this.currentPosition + this.readCount;
	}

	@Override
	public void resetPosition() {
		if(this.readLimit > 0) {
			this.currentPosition = 0 - this.readLimit;
		} else this.currentPosition = 0;
	}
	// END EXTENSION
	/**
	 * 
	 */
	private void doFill() throws IOException {
		this.head = 0;
		int bufferLength = this.buffer.length;
		if (this.readLimit > 0) {
			int unreadCount = this.available();
			int length = unreadCount > bufferLength ? bufferLength : unreadCount;
			this.tail = waitReadBuffer(length);
		} else {
			this.tail = this.is.read(this.buffer, 0, this.buffer.length);
		}
		if(this.tail <= 0) throw new EOBinlogException(EOBinlogException.END_OF_BINLOG_IS_REACHED);
	}

	public int waitAvailable() throws IOException {
		int available = available();
		if (available <= 0) waitAvailable(0);
		return available();
	}

	private void waitAvailable(int readBytes) throws IOException {
		if (is.available() == 0) {
			long readStartTime = System.currentTimeMillis();
			long delay = -1;

			// Sleep for up to timeout.
			while (is.available() == 0) {
				try {
					Thread.sleep(10);
				}
				catch (InterruptedException e) {
					throw new IOException("Reading of event packet was interrupted");
				}
				delay = System.currentTimeMillis() - readStartTime;
				if (delay > timeoutMillis) {
					break;
				}
			}

			// Note the delay if longer than 10% of the timeout. This
			// is helpful for diagnosing failures.
			if (delay > (timeoutMillis / 10)) {
				LOGGER.info("Paused to allow packet data to appear on the network:"
						+ " delay=" + (delay / 1000.0)
						+ " timeout=" + (timeoutMillis / 1000.0)
						+ (packetNumber > 0 ? " packetNumber=" + packetNumber : " positioning or skipping")
						+ " packetlen=" + readLimit
						+ " readBytes=" + readBytes);
			}
		}
	}

	private int waitReadBuffer(int length) throws IOException {
		int n = 0;
		while (n < length) {
			waitAvailable(n);

			// Now read data.
			int count = is.read(this.buffer, n, length - n > is.available() ? is.available() : length - n);

			boolean emptySocket = (count == 0 && is.available() == 0);
			if (count < 0 || emptySocket) {
				if (transportCallback != null) transportCallback.disconnect();
				if (emptySocket)
					throw new EOBinlogException("Input stream is empty: available bytes = 0");
				throw new EOBinlogException(
						"Reached end of input stream:"
								+ (packetNumber != 0 ? " packetNumber=" + packetNumber : "")
								+ " packetlen=" + readLimit
								+ " bytesRead=" + n);
			}

			n += count;
		}

		return n;
	}
	
	private long doSkip(final long n, boolean waitRead) throws IOException {
		long total = n;
		while(total > 0) {
			final int available = this.tail - this.head;
			if(available >= total) {
				this.head += total;
				break;
			} else {
				total -= available;
				if (waitRead) {
					int bufferLength = this.buffer.length;
					int length = total > (long)bufferLength ? bufferLength : (int)total;
					setReadLimit(length);
					doFill();
					setReadLimit(0);
				} else doFill();
			}
		}
		return n;
	}
	
	private int doRead(final byte[] b, final int off, final int len) throws IOException {
		int total = len;
		int index = off;
		while(total > 0) {
			final int available = this.tail - this.head;
			if(available >= total) {
				System.arraycopy(this.buffer, this.head, b, index, total);
				this.head += total;
				break;
			} else {
				System.arraycopy(this.buffer, this.head, b, index, available);
				index += available;
				total -= available;
				doFill();
			}
		}
		return len;
	}

	public void setTimeoutMillis(long timeoutMillis) {
		this.timeoutMillis = timeoutMillis;
	}
}
